; Communication routines
; coded for GPU for faster communication
; based on BJL by 42Bastian Schick

*
* This little routine loads data via the joyport.
* Input : Destination (long)
*         Length in bytes (long)
*         Data (long)
* Output: Data stored from Destination ...
*
* Note : BUSY is high active , /STROBE is low-active.
*        Like your printer wants it ;-)
*

	GPU			; don't forget this!!!
	org $f03000

PACKET_SIZE	equ 4096		; packet size in bytes + checksum

lastPacketSize	reg r17
packetSize	reg r18
dst			reg r19
GetLong		reg r20
SendBool	reg r21
checksum        REG r22
bytesRemaining	reg r23		; bytes to upload
bytesToGo		reg r24		; address of shared variable
return reg r25
help_ptr        REG r26
Magic           REG r27

mainGPU
	movei #ReceiveLong,GetLong
	movei #SendBit,SendBool
	movei #$3ffc,bytesToGo
	moveq #0,checksum
	moveq #0,bytesRemaining
	store bytesRemaining,(bytesToGo)

	movei #.cont0,return
	moveq #0,r0                     ; init r0 to get FALSE
	movei #$22071970,Magic

.cont0
	cmp r0,Magic                    ; magic ??
	jump nz,(GetLong)               ; no, fetch next long
	nop

	movei #.cont1,return
	jump (GetLong)                  ; get start
	moveq #0,help_ptr

.cont1
	jump z,(GetLong)                ; again , if timeout
	store r0,(help_ptr)             ; save start for 68k-routine

	movei #.cont2,return
	jump (GetLong)                  ; get length in bytes
	move r0,dst                ; destination = start address

.cont2
	jump z,(GetLong)                ; again, if timeout
	move r0,bytesRemaining

packetLoop
	moveq #0,checksum               ; clear checksum
	movei #PACKET_SIZE,packetSize
	cmp packetSize,bytesRemaining	; bytesRemaining < PACKET_SIZE
	jr pl,.mayor
	nop
	move bytesRemaining,packetSize	; for small programs & last packet
.mayor
	move packetSize,lastPacketSize
	addq #4,packetSize				; add checksum size
	movei #.byteReaded,return
.bytesLoop
	jump (GetLong)                  	; get next long
	store bytesRemaining,(bytesToGo)	; some feedback
	jr z,.bytesLoop						; loop if timeout
	nop
.byteReaded
	add r0,checksum						; a simple checksum
	subq #4,bytesRemaining				; dec counter
	subq #4,packetSize					; dec counter
	store r0,(dst)						; save long
	jr nz,.bytesLoop					; loop if packetSize > 0
	addqt #4,dst						; inc ptr

	subqt #4,dst						; move write pointer 1 long back
	addqt #4,bytesRemaining				; inc counter

	cmpq #0,checksum					; sum must be zero
	jr z,.sendOk
	nop
; send fail
	movei #packetLoop,return
	sub lastPacketSize,dst				; move back write pointer
	add lastPacketSize,bytesRemaining	; add packet size
	jump (SendBool)
	moveq #1,r0							; 1 = fail

.sendOk
	movei #.moreData,return
	jump (SendBool)
	moveq #0,r0							; 0 = ok
.moreData
	movei #packetLoop,return
	cmpq #0,bytesRemaining
	jump nz,(return)
	nop

; stop gpu
	movei #$f02114,r0		; GPU control
	load (r0),r1			; read state
	bclr #0,r1				; diseable go flag
.wait_gpu
	store r1,(r0)			; write new state
	jr .wait_gpu			; until stop
	nop
; now the 68000 will continue

value		reg r0		; long read
joystick	reg r1
busyHigh	reg r2
busyLow		reg r3
counter		reg r4
loop		reg r5
dataPort	reg r6
TIMEOUT		reg r8
dummy		reg r9
timeOutCnt	reg r10

ReceiveLong
; init part
	movei #$f14000,joystick
	movei #$81f0,busyHigh
	movei #$8170,busyLow
	moveq #0,value

	movei #TimeOut,TIMEOUT

	movei #$100,dummy                  ; switch port to input

	movei #20000,timeOutCnt
	storew busyLow,(joystick)       ; BUSY inactive

	moveq #4,counter				; 4 bytes
	movei #.byteLoop,loop
.byteLoop
	subq #1,timeOutCnt
.wait1
	load (joystick),dataPort
	jump z,(TIMEOUT)
	btst #3,dataPort
	jr nz,.wait1                     ; wait for /STROBE low
	subq #1,timeOutCnt

	storew dummy,(joystick)			; switch to input
									; set also BUSY = high
; Note: This works only, because the input-lines are pulled-up to 5V.

	load (joystick),r12             ; get data-bits 2..0

; bits in dataPort
;     pos 31 30 29 28 27 26 25 24 23 22 21 20 19 18 17 16 15 14 13 12 11 10 9 8 7 6 5 4 3 2 1 0
;    data  7  6  5  4  x  x  x  x B  2   1  0  x  x  x  x  x x   x  x  x  x x x x x x x S 3 x x
; B = BUSY (output)  high-active
; S = /STROBE (input) low-active

	move dataPort,r13
	shrq #2,dataPort               ; bit 3     => 0
	shrq #28,r13                    ; bits 7..4 => 3..0
	shlq #31,dataPort              ; bit 3     => 31
	shlq #31-22,r12                 ; bits 2..0 => 31..29
	shlq #4,r13                     ; bits 7..4 => 7..4
	shrq #28,dataPort              ; bit 3     => 3
	or r13,value                    ; value |= (7..4)
	shrq #31-2,r12                  ; bits 2..0 => 2..0
	or dataPort,value              ; value |= (3)
	or r12,value                    ; value |= (2..0)

	subq #1,timeOutCnt
.wait2
	load (joystick),dataPort
	jump z,(TIMEOUT)
	btst #3,dataPort
	jr z,.wait2                    ; wait for /STROBE high
	subq #1,timeOutCnt

	subq #1,counter
	storew busyLow,(joystick)		; set busy low
	jump nz,(loop)					; 
	rorq #8,value					; rotate value and read next byte

	xor dataPort,dataPort
	storew busyHigh,(joystick)
	jump (return)
	subq #1,dataPort               ; clear Z-flag

TimeOut:
	storew busyHigh,(joystick)
	jump (return)
	xor dataPort,dataPort         ; set Z-flag

; Protocol to send a byte jaguar to pc (bit by bit)
;
;+--------------------------------------------------------------------+
;|        Jaguar side              |            PC side               |
;+---------------------------------+----------------------------------+
;| Set BUSY low                    | Wait for BUSY to be low          | 
;| Wait for /STROBE to be low      | Set /STROBE to low               |
;|                                 | delay                            |
;| Set BUSY equal high bit         | Read BUSY bit                    |
;|                                 | delay                            |
;| Wait for /STROBE to be high     | Set /STROBE to high              |
;|                                 | delay                            |
;| Loop for next bit               | Loop for next bit                |
;+--------------------------------------------------------------------+
;
; routines are named from jaguar side, this means that SendByte, sends
; a byte from Jaguar to PC

; send a boolean (1 bit) value to pc
; r0=0 send 0
; r0!=0 send 1
SendBit
	movei #$f14000,joystick
	movei #$81f0,busyHigh
	movei #$8170,busyLow

.waitStrobeLow
	load (joystick),dataPort
	btst #3,dataPort
	jr nz,.waitStrobeLow			; wait until pc is ready to read
	storew busyLow,(joystick)		; tell we are ready to send

	cmpq #0,r0
	jr nz,.send_zero
	move busyHigh,value
	jr t,.skip
	nop
.send_zero
	move busyLow,value
.skip

	storew value,(joystick)		; send bit
.waitStrobeHigh
	load (joystick),dataPort
	btst #3,dataPort
	jr z,.waitStrobeHigh		; wait until strobe is high (=> bit readed)
	storew value,(joystick)		; send bit

;	storew busyLow,(joystick)		;
;	storew busyHigh,(joystick)		;
;	storew busyLow,(joystick)		;

	jump (return)
;	storew busyHigh,(joystick)		; for don't overtake pc to jaguar
	nop

	unreg value
	unreg joystick
	unreg busyHigh
	unreg busyLow
	unreg counter
	unreg loop
	unreg dataPort
	unreg dummy
	unreg TIMEOUT
	unreg timeOutCnt
