/*
	BJL-uploader for Linux
	Copyright (c) 1997, Joe Britt
	
	Freely distributable and modifiable.
	I only ask that if you use this as the basis of some other work,
	you give me credit for this original version.  Thanks.

	Special thanks to 42Bastian, for being extremely helpful in debugging
	the communication with the BJL-ROM in the Jag.  He's the man!

		JOE       Joe Britt
		42BS      42Bastian Schick <elw5basc(at)gp.fht-esslingen.de>
		JBS       John Sohn (jsohn@ee.net)
		RJ	    Robert Jurziga <moveq@gmx.net>
		
	Change History:

	30 Oct 97 JOE     First version.
	31 Oct 97 42BS    added DJGPP support, but don't know if it works...
	 1 Dec 97 JBS     added Win95 console support and enhanced DGJPP functionality
	10 Jan 98 JBS     supports linking with Windows 95 GUI application
	14 Feb 98 42BS    GEMDOS - relocator debuged / trans-routines in bjl.s

         4 Jan 99 42BS    removed win95 stuff, tested under Linux.
                          repl. gWait by port-accesse (out is 1usec, ever)
 
          7 Jul 02 RJ	 	Port to BeOS.
         				BJL Driver - A BeOS driver that covers the BJL protocol (sending routines written by 42BS)
         				BJL LIB	- a C++ class library
         				Loader	- Commandline tool that uses BJL Lib and BJL Driver
         				gWait is a bigtime_t value used with the BeOS function "snooze" to let
         				the sending thread sleep for one µs. gWait is part of the BJL Driver.
*/


/*
from http://www.paranoia.com/~filipg/HTML/LINK/PORTS/F_PARALLEL5.html#PARALLEL_008

STATUS
	7 6 5 4 3 2 1 0
	* . . . . . . .  Busy . . (pin 11), high=0, low=1 (inverted)
	. * . . . . . .  Ack  . . (pin 10), high=1, low=0 (true)
	. . * . . . . .  No paper (pin 12), high=1, low=0 (true)
	. . . * . . . .  Selected (pin 13), high=1, low=0 (true)
	. . . . * . . .  Error. . (pin 15), high=1, low=0 (true)
	. . . . . * * *  Undefined

CONTROL
	7 6 5 4 3 2 1 0
	* * . . . . . .  Unused (undefined on read, ignored on write)
	. . * . . . . .  Bidirectional control, see below
	. . . * . . . .  Interrupt control, 1=enable, 0=disable
	. . . . * . . .  Select . . (pin 17), 1=low, 0=high (inverted)
	. . . . . * . .  Initialize (pin 16), 1=high, 0=low (true)
	. . . . . . * .  Auto Feed  (pin 14), 1=low, 0=high (inverted)
	. . . . . . . *  Strobe . . (pin 1),  1=low, 0=high (inverted)
*/

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// I N C L U D E
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
#include "../bjllib/bjlclass.h"
#include <ctype.h>
#include <stdio.h>
#include <iostream.h>
#include <File.h>
#include <ByteOrder.h>

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// P R O T O T Y P E S
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
bool 			ParseArgs		( uint32 argc, char *argv[] );
void			PrintUSAGE	( void );
uint32		getValue 		(char *value);
uint32		relocate		(uint8 *pdata, uint32 *len , uint32 *addr, uint32 skip);


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// D A T A
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
Bjl *BJLCon;
uint32 gHeaderSkip	= 0;
uint32 gBaseAddr   	= 0x4000;
uint32 gBasePort    	= 0x378;

BFile *uploadfile;
char FileName[256];



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// MAIN
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
int main (int argc, char *argv[])
{
	IOCtlData tiod;
	off_t filesize=0;
	uint8 *filemem=0;
		
	cout << "loader - Atari Jaguar BJL uploader tool." << "\n";

	if (!(ParseArgs( argc, argv))) 
		PrintUSAGE();


	// open file to upload
	// ------------------------------------------------------------------
	uploadfile = new BFile (FileName, B_READ_ONLY);
	if (uploadfile->InitCheck() != B_OK) {
		cerr<< "Failt to open upload file !!" << "\n";
		exit (0);
		}
		
	// get file size
	uploadfile->GetSize (&filesize);
	
	// allocate memory for file with aliged to long size
	filemem = new uint8 [(filesize+4) & -3];
	if (filemem == 0) {
		cerr << "COULDN'T ALLOCATE MEMORY!!! \n";
		exit (0);
		}

	// load upload file into memory and relocate if neccassary
	uploadfile->Read( filemem, filesize );
  	gHeaderSkip = relocate(filemem, (uint32 *)&filesize, &gBaseAddr, gHeaderSkip);    // relocate program
	

	// open BJL driver.
	BJLCon = new Bjl();
	BJLCon->Open();
	if (! ( BJLCon->isReady () ) ) {
		cerr << "BJL diriver not installed !!!"<<"\n";
		exit (0);
		}


	// setup Configuration structure and configure BJL
	// ------------------------------------------------------------------
	tiod.Port 		= gBasePort;
	tiod.buffer		= filemem+ gHeaderSkip;
	tiod.bufferSize	= (filesize+4) & -3;			// align to long
	tiod.addr		= gBaseAddr;
	BJLCon->Configure (&tiod);	

	// send file to Jaguar
	cout << "Sending '" << FileName << "'  Size: " << filesize << " \n";
	BJLCon->Send();
		
		
	// free resources and leave
	delete ( filemem );
	BJLCon->Close();
	cout << "Ready ...\n";		
	return (0);
}


/* = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = 
** Function:	PrintUSAGE
**
*/
void PrintUSAGE (void) 
{
	cout << "\n" << "USAGE: loader  [-s skipbytes] [-b baseaddr] [-p lptbase] <filename>\n";
	cout << "\tskipbytes \t # bytes in header to skip.  default = "<< gHeaderSkip << "\n";
	cout << "\tbaseaddr \t upload address . default = $" <<  hex <<  gBaseAddr << "\n";
	cout << "\tlptbase \t specifies base address of LPT port. default = $" << hex << gBasePort << "\n";
	exit (0);
}





/* = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = 
** Function:	ParseArgs
**
** Return:		bool
**
*/

bool ParseArgs ( uint32 n, char *argv[] )
{
	char *cptr;
	
	if (n<=1)
		return (false);
	
	FileName[0]=0x0;
	
	for (uint32 i=1; i < n; ) {
		cptr = argv[i];
		if (*cptr == '-') {
			switch (*(cptr+1)) {
				case 'S':
				case 's':
					if (argv[i+1] != 0) {
						gHeaderSkip = getValue ( argv[i+1] );
						i+=2;
						}
					else
						return ( false );
					break;
				
				case 'B':
				case 'b':
					if (argv[i+1] !=0) {
						gBaseAddr = getValue (argv[i+1]);
						i+=2;
						}
					else
						return ( false );
					break;
				
				case 'P':
				case 'p':
					if ( argv [i+1] != 0) {
						gBasePort = getValue (argv[i+1]);
						i+=2;
						}
					else
						return ( false );
					break;
				
				
				default:
					cerr << "SYNTAX ERROR!" << "\n";
					return ( false );
				}
			}
		else {
			strcpy (&FileName[0], argv[i]);
			i++;
			}
		}	

	if (FileName[0] == 0x0) {
		cerr << "SYNTAX ERROR: Missing file to upload!" << "\n";
		return ( false );
		}
		
	#ifdef DEBUG
		cout << "SkipBytes: "<< gHeaderSkip << "\n";
		cout << "Basis Upload Address :$"<< hex << gBaseAddr << "\n";
		cout << "Port Base :$" << hex << gBasePort << "\n";
		cout << "File to load:" << FileName << "\n";
	#endif
	
	return ( true );
}




/* = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = 
** Function:	getValue
**
** Return:		uint32
**
*/

uint32 getValue (char *value)
{
	uint32 retval = 0;
	if (*value == 'x') 
		sscanf ( value+1 , "%lx", &retval);
	else
		sscanf (value, "%ld", &retval);
	return ( retval );
}






/**********************/
/* GEMDOS - relocator */
/**********************/
/* this function sees if it finds a GEMDOS or JServer - header
	 if so, skip,addr and len are adjusted
*/

uint32 relocate( uint8 *pdata, uint32 *len , uint32 *addr, uint32 skip)
{

  if ( (pdata[0] == 0x60) && (pdata[1] == 0x1a) ){ // GEMDOS - header 

    if ( *((ulong *)(pdata+28)) == B_BENDIAN_TO_HOST_INT32 (0x4a414752) ){ /* JAGR */
      
      ushort type = B_BENDIAN_TO_HOST_INT16 (*((ushort *)(pdata+32)));
      ulong new_len;
	
      if ( type != 2 && type != 3 ){
	
	cerr << "Unsupported JServer-Header!\nLeaving ..\n";
	
//	 printf("Unsupported JServer-Header !\nLeaving ...\n");
	exit(-1);
      }
		
      *addr = B_BENDIAN_TO_HOST_INT32 ( *(ulong *)(pdata+34) );  // dest and start address
      
      if ( (new_len = B_BENDIAN_TO_HOST_INT32 ( *(ulong *)(pdata+38) ) )){
	
	*len = new_len;
      }
      
      if ( type == 2)
	return (28+14);
      else
	return (28+18);   // skip GEMDOS and JServer-header
    } else {
//
// BJL - program
// 
      ulong textlen,datalen,bss_len,symblen;   /* header info */
      ulong offset;
      ulong addr2 = *addr;
      uchar * reloptr, * pdata2;
      
//    printf("Relocating BJL ...\n");

      textlen = B_BENDIAN_TO_HOST_INT32 (*(ulong *)(pdata+2));
      datalen = B_BENDIAN_TO_HOST_INT32 (*(ulong *)(pdata+6));
      bss_len = B_BENDIAN_TO_HOST_INT32 (*(ulong *)(pdata+10)); //  no need for this, yet
      symblen = B_BENDIAN_TO_HOST_INT32 (*(ulong *)(pdata+14));
      
//    printf("TEXT(%08x) DATA(%08x) BSS(%08x) SYM(%d)\n",textlen,datalen,bss_len,symblen);

      reloptr = pdata + textlen + datalen + symblen + 28;

      offset = B_BENDIAN_TO_HOST_INT32 (*(ulong *)reloptr);   // first long to relocate

//    printf("Offset :%08x\n",offset);

      reloptr += 4;                 // from now only bytes as offset
		
      pdata2 = pdata + 28 + offset;      // ptr to it

      addr2 = B_BENDIAN_TO_HOST_INT32 (addr2);        // change byte-order 
		
      while ( offset ){              // until offset == 0	
      // compute new address 
//      *(ulong *)pdata2 = LE_TO_BE(BE_TO_LE( *(ulong *)pdata2 ) + addr2);
	*(ulong *)pdata2 += addr2;       // ha, it works ;-)

      // as long as offset is 1 skip 254 byte
	do{
      
	  offset = (ulong) *reloptr;
	  ++reloptr;
	  if (! --offset)
	    pdata2 += 254;
	}while ( !offset );
	pdata2 += ++offset;           // next address
      }

      *len = textlen + datalen;     // send only text and data , no symbols ...
      return ( 28 );                // skip GEMDOS header
    }
  } else if ( (pdata[0] == 0x01) && (pdata[1] == 0x50) ) {

//    printf("COFF-Header found ...\n");
    cout << "COFF-Header found ...\n";					// RJ
    
    *addr = B_BENDIAN_TO_HOST_INT32 (*(ulong *)(pdata+56));

    skip = B_BENDIAN_TO_HOST_INT32 (*(ulong *)(pdata+68));

    *len -= skip;

    return skip;

  } else {
  
//    printf("No Gemdos or COFF header  !\n"); 	
    cout << "No Gemdos or COFF header !\n";				// RJ
    
    return (skip);                    /* skip user-defined bytes */
  }
}

